import numpy as np


# --------------------------------------------------------------------------
#            Diagnostics
# --------------------------------------------------------------------------
# Bit of a placeholder for differencing tests

def sign_changes(lagged):
    return np.nansum([abs(d) > 1.5 for d in np.diff(np.sign(list(lagged) + [0., 0.]))])


def is_process(lagged):
    return sign_changes(np.diff(lagged)) > 2 * sign_changes(lagged)


# --------------------------------------------------------------------------
#            Monte Carlo
# --------------------------------------------------------------------------

def inv_cdf_walk(inv_cdf, k: int, x0: float) -> float:
    """ Generate sample from sum of independent increments generated by inverse cdf function
         inv_cdf  func   taking (0,1) --> R
         k        int    Number of steps
         x0       float  Starting value
    """
    # See crawler_examples/decastyle_cat for usage example
    x = x0
    for _ in range(k):
        u = np.random.rand()
        dx = inv_cdf(u)
        x = x + dx
    return x




from collections import OrderedDict
from microprediction.crawler import MicroCrawler


class StreamCrawler(MicroCrawler):

    # Crawler maintaining state for each stream
    # See examples/thallodal_cat.py for an illustration of use

    ########################################################
    #  Override the following methods                      #
    ########################################################

    def initial_state(self, name, lagged_values, lagged_times, **ignore):
        return None

    def update_state(self, state, lagged_values=None, lagged_times=None, **ignore):
        return state

    def sample_using_state(self, state, lagged_values, lagged_times, name, delay, **ignore):
        raise NotImplemented('Need to implement sample_using_state when deriving from StreamCrawler')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.stream_state = OrderedDict()

    ########################################################
    #  Maybe leave this alone                              #
    ########################################################

    def sample(self, lagged_values, lagged_times=None, name=None, delay=None, **ignored):
        """ Calls sample_using_state """

        if name not in self.stream_state:
            self.stream_state.update(
                {name: self.initial_state(name=name, lagged_values=lagged_values, lagged_times=lagged_times)})
        state = self.stream_state[name]
        state = self.update_state(state=state, lagged_values=lagged_values, lagged_times=lagged_times)
        samples = self.sample_using_state(state=state, lagged_values=lagged_values, lagged_times=lagged_times,
                                          name=name, delay=delay)
        self.stream_state[name] = state
        return samples











# Default crawler
# This may change from time to time.
# We do not recommend deriving from DefaultCrawler

from microprediction import SequentialStreamCrawler
DefaultCrawler = SequentialStreamCrawler




from abc import ABC
from copy import deepcopy
import numpy as np
from collections import OrderedDict


class DistMachine(object):

    # This is a state engine that has a ppf function
    # Example of usage: SequentialStreamCrawler

    def __init__(self, state, params: OrderedDict = None):
        self.params = params
        self.state = state

    def update(self, value=None, dt=None, **kwargs):
        # Incorporate new value, time passing, or both
        # Typically will update the state but not params
        raise NotImplementedError

    def inv_cdf(self, p: float) -> float:
        # Something like StatsConventions.norminv(p)
        raise NotImplementedError


class LossDist(DistMachine):

    # A distribution machine with a loss function

    def __init__(self, state: dict, params: OrderedDict = None):
        super().__init__(state=state, params=params)

    def log_likelihood(self, value: float) -> float:
        """ Likelihood for one value """
        raise NotImplementedError

    # Likelihood Distribution Machine comes with a default loss function, though you could
    # choose to augment or override this by adding other penalties (for instance adding penalty
    # for moving too much)

    def loss(self, lagged_values, lagged_times, params):
        """ Loss function for a series of values """
        # Lower the better
        saved_params = deepcopy(self.params)
        self.params = deepcopy(params)
        chronological_values = list(reversed(lagged_values))
        chronological_times = list(reversed(lagged_times))
        chronological_dt = [1.0] + list(np.diff(chronological_times))

        ll = 0.0
        for value, dt in zip(chronological_values, chronological_dt):
            self.update(dt=dt)
            ll += -self.log_likelihood(value=value)
            self.update(value=value)
        self.params = saved_params
        return ll
